      PROGRAM4
      USE numerical_libraries
C*********************************************************************
c     Program 4 of Subba Rao -- Gabr MONOGRAPH:
C         "AN INTRODUCTION TO BISPECTRAL ANALYSIS AND BILINEAR TIME      
C         SERIES MODELS"  (SPRINGER-VERLAG, 1984)
C         Version: May 4, 2012
C
c     The program estimates the bispectral density function and 
C     the spectral density function f(\omga_i). The null hypothesis H_0: 
C
c        |f(\omega_i,\omega_j)|^2/f(\omega_i)f(\omega_j)f(\omega_i+\omega_j)
c        = constant, is tested
C                                                                     
C     ORIGINAL NAG ROUTINE:
C         F04ASF calculates the accurate solution of a set of real  
C         symmetric positive-definite linear equations with
C         a single right-hand side, Ax = b, using a Cholesky factorization 
C         and iterative refinement.
C     REPLACED BY 2 IMSL ROUTINES:
C         LFCDS Compute the R^TR Cholesky factorization of a real symmetric
C               positive definite matrix and estimate its L_1 condition 
C               number
C         LFIDS use iterative refinement to improve the solution of a real
C               symmetric positive definite system of linear equations
C
C     N = NUMER OF OBSERVATIONS
C     M = NUMBER OF THIRD ORDER COVARIANCES USED IN THE ESTIMATION OF THE
C         BISPECTRAL DENSITY FUNCTION
C     
C     NP = K = [4,5,6,7,8]
C     K5 = K
c     IW = 1 Bartlett window; 2 Tukey window; 3 Parzen window
C
C*********************************************************************
      INTEGER I,KK,I11,I12,I2,LDA,LDFAC,L2
      PARAMETER (LDA=6,LDFAC=6,NN=6)  ! REPLACE IF APPROPRIATE
      REAL FACC(LDFAC,LDFAC),COND,RES(NN),AINV(LDA,LDA)
      
      COMMON/A1/W(10,12)
      COMMON/A2/C(72,72)
      COMMON/A3/V(72)
		
      COMMON/A4/FM(8,8,17),FN(8,8,17)
      COMMON/A5/X(1202),IR(1202),RU(1202),E(1202)
      COMMON/A6/WK1(15),WK2(15)
      COMMON/A7/F(15,17),FAVR(15),Z(6,6),Y(15),FT(17,15)
      COMMON/A8/R(72)

      N=114
      IW=3
      M=16
      NP=6    ! K IN Monograph 
      K5 = 4  ! 5, 6, 7, 8
      
      KS=K5

      OPEN(7,FILE='C:\...\LYNXN114.DAT',STATUS='UNKNOWN')
          READ(7,*) (X(I),i=1,n)
   
      SUMX=0.
      DO 90 I=1,N
90       SUMX=SUMX+X(I)
  	
      XMEAN=SUMX/FLOAT(N)
      DO 140 I=1,N
140      X(I)=X(I)-XMEAN
      M4=M+1
      DO 10 J2=1,M4
  	 DO 10 J1=J2,M4
  	    SUM=0
  	    Nl=N-J1+1
  	    DO 20 K=1,NL
  	       K1=J1-1+K
  	       K2=J2-1+K
20             SUM=SUM+X(K1)*X(K2)*X(K)
	       C(J1,J2)=SUM/FLOAT(N)
10       CONTINUE
  	 CALL WINDOW(M,IW)
  	 DO 15 I=1,M4
  	    SUM=0.
  	    NI=N-I+1
  	    DO 16 J=1,NI
16             SUM=SUM+X(J)*X(I+J-1)
15             R(I)=V(I)*SUM/FLOAT(N) ! FLOAT(N-I) REPLACED BY FLOAT(N)
      PI=4.*ATAN(1.)
      KF=2*KS+1
      NP1=NP+1
      NP2=NP-1
      SN=FLOAT(N)/FlOAT(NP)
      PK=6                   ! IN MONOGRAPH 3.5, SUBSTITUTED FOR D 

      DO 40 I=1,NP2
  	   DO 40 KK=1,KF 
  	      W(I,KK)=(FlOAT(I)*SN+FLOAT(KK-KS-1)*PK)/FLOAT(N)
40    CONTINUE
  	G=(1./(2.*PI))**2
  	LP=2*NP/3
  	DO 60 I1=1,LP
  	   V1=W(I1,KS+1)

	   D3=B(V1)
  	   I12=NP-I1/2-1
  	   I11=I1+1
  	   IF (I11.GT.I12) GOTO 61
  	   DO 60 I2=I11,I12
  	      U2=W(I2,KS+1)
	      D2=B(U2)
  	      DO 60 I3=1,KF
  	         U1=W(I1,I3)
  	         V2=W(I2,I3)
                 ARF2=C(1,1)
             	 ARF1=C(1,1)
  	         AIF1=0  ! X MISTAKE 
  	         AIF2=0
  	         DO 50 J2=2,M
  	            JJ=J2+1
  	            DO 50 J1=JJ,M4
  	               S=V(J1)*V(J2)*V(J1-J2+1)*C(J1,J2)  ! (J1-J2) MODIFIED
      A1=FLOAT(J1-1)*PI
      A2=FLOAT(J2-1)*PI
      A12=A1-A2
      ARF1=ARF1+S*(COS(A1*U1+A2*U2)+COS(A1*U2+A2*U1)+COS(A1*U1+A12*U2)
     *+COS(A1*U2+A12*U1)+COS(A12*U1-A2*U2)+COS(A12*U2-A2*U1))
     
      ARF2=ARF2+S*(COS(A1*V1+A2*V2)+COS(A1*V2+A2*V1)+COS(A1*V1+A12*V2)
     *+COS(A1*V2+A12*V1)+COS(A12*V1-A2*V2)+COS(A12*v2-A2*V1))
     
      AIF1=AIF1-S*(SIN(A1*U1+A2*U2)+SIN(A1*U2+A2*U1)-SIN(A1*U1+A12*U2 ) 
     *-SIN(A1*U2+A12*U1)+SIN(A12*U1-A2*U2)+SIN(A12*U2-A2*U1))
     
      AIF2=AIF2-S*(SIN(A1*V1+A2*V2)+SIN(A1*V2+A2*V1)-SIN(A1*V1+A12*V2)
     *-SIN(A1*V2+A12*V1)+SIN(A12*V1-A2*V2)+SIN(A12*V2-A2*V1))
50    CONTINUE
 
      DO 51 I=2,M4
  	   AAI=FLOAT(I-1)*PI
  	   S=(C(I,1)+C(I,I))*(V(I)**2)
   	   H=(C(I,1)-C(I,I))*(V(I)**2)
   	   ARF1=ARF1+S*(COS(AAI*U1)+COS(AAI*U2)+COS(AAI*(U1+U2)))
   	   ARF2=ARF2+S*(COS(AAI*V1)+COS(AAI*V2)+COS(AAI*(V1+V2)))
   	   AIF1=AIF1+H*(SIN(AAI*(U1+U2))-SIN(AAI*U1)-SIN(AAI*U2))
   	   AIF2=AIF2+H*(SIN(AAI*(V1+V2))-SIN(AAI*V1)-SIN(AAI*V2))
51    CONTINUE
      FM(I1,I2,I3)=G**2*(ARF1**2+AIF1**2)
      FN(I1,I2,I3)=FM(I1,I2,I3)/(D2*B(U1)*B(U1+U2))

      IF (I3-K5-1) 57,60,52
57    J3=I3+KF

      GOTO 53
      
52    J3=I3+KF-1


53    FM(I1,I2,J3)=G**2*(ARF2**2+AIF2**2)
      FN(I1,I2,J3)=FM(I1,I2,J3)/(D3*B(V2)*B(V1+V2))
      
60    CONTINUE
  
61    CONTINUE

       WRITE(*,*) 'IWINDOW (IW) =', IW
       IF(IW-2) 501,502,503

501     CONTINUE
   	  GOTO 504
502     CONTINUE
          GOTO 504
503     CONTINUE
        
   	NPL=NP-LP-1	

504   CONTINUE    	
      DO 300 L2=1,NPL  
   	   I2=NP-L2	
   	   I21=2*L2-1	
	   
300    CONTINUE	 
   	NL1=LP-1
   	DO 400 L2=1,NL1
   	   I2=LP-L2+1
   	   I21=I2-1   
400   CONTINUE
   	DO 310 L2=1,NPL
   	   I2=NP-L2
   	   I21=2*L2-1
310   CONTINUE
   	DO 410 L2=1,NL1
   	   I2=LP-L2+1
   	   I21=I2-1
410   CONTINUE
   	KFF=2*KF-1
   	DO 560 I1=1,LP
           I12=NP-I1/2-1
           I11=I1+1
           IF(I11.GT.I12) GOTO 561
           DO 560 I2=I11,I12
560   CONTINUE

561   CONTINUE
   	DO 570 I1=1,LP
   	   I12=NP-I1/2-1
   	   I11=I1+1
   	   IF(I11.GT.I12)GOTO 571	
   	   DO 570 I2=I11,I12
570        CONTINUE
571   L=0
   	DO 660 I1=1,LP
   	   I12=NP-I1/2-1
   	   I11=I1+1
   	   IF(I11.GT.I12)GOTO 662
   	   DO 660 I2=I11,I12
   	      L=L+1 
  	      DO 660 K=1,KFF
    	         F(L,K)=FN(I1,I2,K)
660     CONTINUE
662   N=7
      K=KFF
      N1=N-1
   	DO 678 I=1,N1
   	DO 678 J=1,K
    	     F(I,J)=F(I,J)-F(I+1,J)
678     CONTINUE
        N=N-1
        DO 8 I=1,N
           FAVR(I)=0.0
           DO 7 J=1,K
   	      FAVR(I)=FAVR(I)+F(I,J)
7          CONTINUE
           FAVR(I)=FAVR(I)/K
8       CONTINUE
   	DO 9 I=1,N
   	DO 9 J=1,K
   	   F(I,J)=F(I,J)-FAVR(I)
   	   FT(J,I)=F(I,J)
9     CONTINUE
   	DO 19 I1=1,N
   	DO 19 I2=I1,N
   	   Z(I1,I2)=.0
   	   DO 19 J=1,K
   	      Z(I1,I2)=Z(I1,I2)+F(I1,J)*FT(J,I2)
   	      Z(I2,I1)=Z(I1,I2)
19    CONTINUE

       CALL LSASF(N,Z,LDA,FAVR,Y)   

   	Q=0.
   	DO 18 I=1,N
   	   Q=Q+(FAVR(I))*Y(I)
18      CONTINUE
   	T=K*Q
   	WRITE(*,*) 'HOTELLINGS T SQUARE STATISTIC IS = ', T
   	STOP
   	END

      SUBROUTINE WINDOW(M,IW)
      COMMON/A3/V(72)
      IF(IW-2)1,2,3
1     CALL BTLETT(M)
      GOTO 4
2     CALL TUKEY(M)
      GOTO 4
3     CALL PARZEN(M)
4     RETURN
      END
        
      SUBROUTINE BTLETT(M)
      COMMON/A3/V(72)
      M4=M+1
      DO 12 I=1,M4
         V(I)=1.-FLOAT(I-1)/FLOAT(M)
12    CONTINUE
      RETURN
      END
        
      SUBROUTINE TUKEY(M)
      COMMON/A3/V(72)
      M4=M+1
      PI=4.*ATAN(1.)
      DO 12 I=1,M4
           V(I)=0.54+0.46*COS(PI*FLOAT(I-1)/FLOAT(M))
12    CONTINUE
      RETURN
      END
     
      SUBROUTINE PARZEN(M)
      COMMON/A3/V(122)
      M4=M+1
      DO 12 I=1,M4
         K=I-1
         H1=FLOAT(K)/FLOAT(M)
         IF(K-M/2)13,13,14
13       V(I)=1.-6.*(H1**2-H1**3)
         GOTO 12
14       V(I)=2.*((1.-H1)**3)
12    CONTINUE
      RETURN
      END
    
   	FUNCTION B(U)
   	COMMON/A8/R(72)
   	M1=29
   	PI=4.*ATAN(1.)
   	SUM=0.0 
   	DO 10 I=1,M1
10         SUM=SUM+R(I+1)*COS(PI*U*FLOAT(I))
   	B=(R(1)+2.*SUM)/(2.*PI)
   	RETURN
   	END
  